import unittest
import tempfile
from gillespy2.core.gillespyError import DirectoryError, SimulationError
from example_models import Example
from gillespy2 import SSACSolver, ODECSolver, TauLeapingCSolver


class TestVariableSolvers(unittest.TestCase):
    model = Example()
    solverSSAC = SSACSolver(model, variable=True)
    solverODEC = ODECSolver(model, variable=True)
    solverTAUC = TauLeapingCSolver(model, variable=True)
    solverlist = [solverSSAC, solverODEC, solverTAUC]

    def test_create(self):
        model = Example()
        solverSSAC = SSACSolver(model)
        solverODEC = ODECSolver(model)
        solverTAUC = TauLeapingCSolver(model)

    def test_file_with_directory_name_exists(self):
        with self.assertRaises(DirectoryError):
            temp = tempfile.NamedTemporaryFile()
            model = Example()
            solverSSAC = SSACSolver(model, temp.name)
            solverODEC = ODECSolver(model, temp.name)
            solverTAUC = TauLeapingCSolver(model, temp.name)

    def test_run_example_precompiled(self):
        for solver in self.solverlist:
            results = self.model.run(solver=solver)

    def test_change_species(self):
        initial_value = self.model.listOfSpecies['Sp'].initial_value
        for solver in self.solverlist:
            results = self.model.run(solver=solver, variables={'Sp':3})
        with self.subTest(msg='Test changed species simulation'):
            self.assertEqual(results['Sp'][0], 3)
        with self.subTest(msg='Test changed species model integrity'):
            self.assertEqual(self.model.listOfSpecies['Sp'].initial_value, initial_value)

    def test_change_parameter(self):
        initial_expression = self.model.listOfParameters['k1'].expression
        for solver in self.solverlist:
            print(solver.name)
            results = self.model.run(solver=solver, variables={'k1':0})
            with self.subTest(msg='Test changed parameter simulation'):
                self.assertEqual(results['Sp'][-1], results['Sp'][0])
            with self.subTest(msg='Test changed parameter model integrity'):
                self.assertEqual(self.model.listOfParameters['k1'].expression, initial_expression)

    def test_invalid_variable(self):
        with self.assertRaises(SimulationError):
            for solver in self.solverlist:
                results = self.model.run(solver=solver, variables={'foobar':0})

    def test_run_example(self):
        notPrecompiled = [SSACSolver, ODECSolver, TauLeapingCSolver]
        for solver in notPrecompiled:
            results = self.model.run(solver=solver)


if __name__ == '__main__':
    unittest.main()




"""
This file contains a function and variable for testing a machines support of GillesPy2 C++ solvers.
Used in model.py
"""

def check_cpp_support():
    import shutil

    dependencies = ['g++', 'make']
    missing = []
    any_missing = False

    for dependency in dependencies:
        if shutil.which(dependency) != None:
            continue

        missing.append(dependency)
        any_missing = True

    if any_missing is True:
        from gillespy2.core import log
        log.warn('Unable to use C++ optimized SSA due to one or more missing dependencies: {0}. '
        'The performance of this package can be significantly increased if you install/configure '
        'these on your machine.'.format(missing))

    return not any_missing

cpp_support = check_cpp_support()



from gillespy2.core.reaction import *
from gillespy2.core.raterule import RateRule
from gillespy2.core.parameter import Parameter
from gillespy2.core.species import Species
from gillespy2.core.reaction import Reaction
import numpy as np
from gillespy2.core.results import Trajectory,Results
from collections import OrderedDict
from gillespy2.core.gillespyError import *

try:
    import lxml.etree as eTree

    no_pretty_print = False

except:
    import xml.etree.ElementTree as eTree
    import xml.dom.minidom
    import re
    no_pretty_print = True


def import_SBML(filename, name=None, gillespy_model=None):
    """
    SBML to GillesPy model converter. NOTE: non-mass-action rates
    in terms of concentrations may not be converted for population
    simulation. Use caution when importing SBML.

    :param filename: Path to the SBML file for conversion.
    :type filename: str

    :param name: Name of the resulting model
    :type name: str

    :param gillespy_model: If desired, the SBML model may be added to an existing GillesPy model
    :type gillespy_model: gillespy.Model
    """

    try:
        from gillespy2.sbml.SBMLimport import convert
    except ImportError:
        raise ImportError('SBML conversion not imported successfully')

    return convert(filename, model_name=name, gillespy_model=gillespy_model)


def export_SBML(gillespy_model, filename=None):
    """
    GillesPy model to SBML converter

    :param gillespy_model: GillesPy model to be converted to SBML
    :type gillespy_model: gillespy.Model

    :param filename: Path to the SBML file for conversion
    :type filename: str
    """
    try:
        from gillespy2.sbml.SBMLexport import export
    except ImportError:
        raise ImportError('SBML export conversion not imported successfully')

    return export(gillespy_model, path=filename)


class Model(SortableObject):
    # reserved names for model species/parameter names, volume, and operators.
    reserved_names = ['vol']
    special_characters = ['[', ']', '+', '-', '*', '/', '.', '^']

    """
    Representation of a well mixed biochemical model. Contains reactions,
    parameters, species.

    :param name: The name of the model, or an annotation describing it.
    :type name: str
    
    :param population: The type of model being described. A discrete stochastic model is a
    population model (True), a deterministic model is a concentration model
    (False). Automatic conversion from population to concentration models
    may be used, by setting the volume parameter.
    Attributes
    :type population: bool

    :param volume: The volume of the system matters when converting to from population to
    concentration form. This will also set a parameter "vol" for use in
    custom (i.e. non-mass-action) propensity functions.  
    :type volume: float   
    
    :param tspan: The timepoints at which the model should be simulated. If None, a
    default timespan is added. May be set later, see Model.timespan
    :type tspan: numpy ndarray
    
    :param annotation: Option further description of model
    :type annotation: str
    """

    def __init__(self, name="", population=True, volume=1.0, tspan=None, annotation="model"):
        """ Create an empty model. """

        # The name that the model is referenced by (should be a String)
        self.name = name
        self.annotation = annotation

        # Dictionaries with model element objects.
        # Model element names are used as keys.
        self.listOfParameters = OrderedDict()
        self.listOfSpecies = OrderedDict()
        self.listOfReactions = OrderedDict()
        self.listOfAssignmentRules = OrderedDict()
        self.listOfRateRules = OrderedDict()
        self.listOfEvents = OrderedDict()
        self.listOfFunctionDefinitions = OrderedDict()

        # Dictionaries with model element objects.
        # Model element names are used as keys, and values are
        # sanitized versions of the names/formulas.
        # These dictionaries contain sanitized values and are for
        # Internal use only
        self._listOfParameters = OrderedDict()
        self._listOfSpecies = OrderedDict()
        self._listOfReactions = OrderedDict()
        self._listOfAssignmentRules = OrderedDict()
        self._listOfRateRules = OrderedDict()
        self._listOfEvents = OrderedDict()
        self._listOfFunctionDefinitions = OrderedDict()
        # This defines the unit system at work for all numbers in the model
        # It should be a logical error to leave this undefined, subclasses
        # should set it
        if population:
            self.units = "population"
        else:
            self.units = "concentration"
            if volume != 1.0:
                raise Warning(
                    "Concentration models account for volume implicitly, explicit volume definition is not required. "
                    "Note: concentration models may only be simulated deterministically.")

        self.volume = volume

        # Dict that holds flattended parameters and species for
        # evaluation of expressions in the scope of the model.
        self.namespace = OrderedDict([])

        if tspan is None:
            self.timespan(np.linspace(0, 20, 401))
        else:
            self.timespan(tspan)

    def __str__(self):
        divider = '\n**********\n'

        def decorate(header):
            return '\n' + divider + header + divider

        print_string = self.name
        if len(self.listOfSpecies):
            print_string += decorate('Species')
            for s in sorted(self.listOfSpecies.values()):
                print_string += '\n' + str(s)
        if len(self.listOfParameters):
            print_string += decorate('Parameters')
            for p in sorted(self.listOfParameters.values()):
                print_string += '\n' + str(p)
        if len(self.listOfReactions):
            print_string += decorate('Reactions')
            for r in sorted(self.listOfReactions.values()):
                print_string += '\n' + str(r)
        if len(self.listOfEvents):
            print_string += decorate('Events')
            for e in sorted(self.listOfEvents.values()):
                print_string += '\n' + str(e)
        if len(self.listOfAssignmentRules):
            print_string += decorate('Assignment Rules')
            for ar in sorted(self.listOfAssignmentRules.values()):
                print_string += '\n' + str(ar)
        if len(self.listOfRateRules):
            print_string += decorate('Rate Rules')
            for rr in sorted(self.listOfRateRules.values()):
                print_string += '\n' + str(rr)
        if len(self.listOfFunctionDefinitions):
            print_string += decorate('Function Definitions')
            for fd in sorted(self.listOfFunctionDefinitions.values()):
                print_string += '\n' + str(fd)

        return print_string

    def serialize(self):
        """ Serializes the Model object to valid StochML. """
        self.resolve_parameters()
        doc = StochMLDocument().from_model(self)
        return doc.to_string()

    def update_namespace(self):
        """ Create a dict with flattened parameter and species objects. """
        self.namespace = OrderedDict([])
        for param in self.listOfParameters:
            self.namespace[param] = self.listOfParameters[param].value

    def sanitized_species_names(self):
        """
        Generate a dictionary mapping user chosen species names to simplified formats which will be used
        later on by GillesPySolvers evaluating reaction propensity functions.
        :return: the dictionary mapping user species names to their internal GillesPy notation.
        """
        species_name_mapping = OrderedDict([])
        for i, name in enumerate(sorted(self.listOfSpecies.keys())):
            species_name_mapping[name] = 'S[{}]'.format(i)
        return species_name_mapping

    def problem_with_name(self, name):
        if name in Model.reserved_names:
            return ModelError(
                'Name "{}" is unavailable. It is reserved for internal GillesPy use. Reserved Names: ({}).'.format(name,
                                                                                                                   Model.reserved_names))
        if name in self.listOfSpecies:
            return ModelError('Name "{}" is unavailable. A species with that name exists.'.format(name))
        if name in self.listOfParameters:
            return ModelError('Name "{}" is unavailable. A parameter with that name exists.'.format(name))
        if name in self.listOfReactions:
            return ModelError('Name "{}" is unavailable. A reaction with that name exists.'.format(name))
        if name in self.listOfEvents:
            return ModelError('Name "{}" is unavailable. An event with that name exists.'.format(name))
        if name in self.listOfRateRules:
            return ModelError('Name "{}" is unavailable. A rate rule with that name exists.'.format(name))
        if name in self.listOfAssignmentRules:
            return ModelError('Name "{}" is unavailable. An assignment rule with that name exists.'.format(name))
        if name in self.listOfFunctionDefinitions:
            return ModelError('Name "{}" is unavailable. A function definition with that name exists.'.format(name))
        if name.isdigit():
            return ModelError('Name "{}" is unavailable. Names must not be numeric strings.'.format(name))
        for special_character in Model.special_characters:
            if special_character in name:
                return ModelError(
                    'Name "{}" is unavailable. Names must not contain special characters: {}.'.format(name,
                                                                                                      Model.special_characters))

    def get_species(self, s_name):
        """
        Returns a species object by name.

        :param s_name: Name of the species object to be returned:
        :type s_name: str
        """
        return self.listOfSpecies[s_name]

    def get_all_species(self):
        """
        Returns a dict of all species in the model, of the form:
        {name : species object}
        """
        return self.listOfSpecies

    def add_species(self, obj):
        """
        Adds a species, or list of species to the model.

        :param obj: The species or list of species to be added to the model object
        :type obj: Species, or list of species
        """

        if isinstance(obj, list):
            for S in sorted(obj):
                self.add_species(S)
        else:
            try:
                problem = self.problem_with_name(obj.name)
                if problem is not None:
                    raise problem
                self.listOfSpecies[obj.name] = obj
                self._listOfSpecies[obj.name] = 'S{}'.format(len(self._listOfSpecies))
            except Exception as e:
                raise ParameterError("Error using {} as a Species. Reason given: {}".format(obj, e))
        return obj

    def delete_species(self, obj):
        """
        Removes a species object by name.

        :param obj: Name of the species object to be removed
        :type obj: str
        """
        self.listOfSpecies.pop(obj)
        self._listOfSpecies.pop(obj)

    def delete_all_species(self):
        """
        Removes all species from the model object.
        """
        self.listOfSpecies.clear()
        self._listOfSpecies.clear()

    def set_units(self, units):
        """
        Sets the units of the model to either "population" or "concentration"

        :param units: Either "population" or "concentration"
        :type units: str
        """
        if units.lower() == 'concentration' or units.lower() == 'population':
            self.units = units.lower()
        else:
            raise ModelError("units must be either concentration or population (case insensitive)")

    def sanitized_parameter_names(self):
        """
        Generate a dictionary mapping user chosen parameter names to simplified formats which will be used
        later on by GillesPySolvers evaluating reaction propensity functions.
        :return: the dictionary mapping user parameter names to their internal GillesPy notation.
        """
        parameter_name_mapping = OrderedDict()
        parameter_name_mapping['vol'] = 'V'
        for i, name in enumerate(sorted(self.listOfParameters.keys())):
            if name not in parameter_name_mapping:
                parameter_name_mapping[name] = 'P{}'.format(i)
        return parameter_name_mapping

    def get_parameter(self, p_name):
        """
        Returns a parameter object by name.

        :param p_name: Name of the parameter object to be returned
        :type p_name: str
        """
        try:
            return self.listOfParameters[p_name]
        except:
            raise ModelError("No parameter named " + p_name)

    def get_all_parameters(self):
        """
        Returns a dict of all parameters in the model, of the form:
        {name : parameter object}
        """
        return self.listOfParameters

    def add_parameter(self, params):
        """
        Adds a parameter, or list of parameters to the model.

        :param params:  The parameter or list of parameters to be added to the model object.
        :type params: Parameter, or list of parameters
        """
        if isinstance(params, list):
            for p in sorted(params):
                self.add_parameter(p)
        else:
            try:
                problem = self.problem_with_name(params.name)
                if problem is not None:
                    raise problem
                self.listOfParameters[params.name] = params
                self._listOfParameters[params.name] = 'P{}'.format(len(self._listOfParameters))
            except Exception as e:
                raise ParameterError("Error using {} as a Parameter. Reason given: {}".format(params, e))
        return params

    def delete_parameter(self, obj):
        """
        Removes a parameter object by name.

        :param obj: Name of the parameter object to be removed
        :type obj: str
        """
        self.listOfParameters.pop(obj)
        self._listOfParameters.pop(obj)

    def set_parameter(self, p_name, expression):
        """
        Set the value of an existing parameter "pname" to "expression".

        :param p_name: Name of the parameter whose value will be set.
        :type p_name: str

        :param expression: String that may be executed in C, describing the value of the
        parameter. May reference other parameters by name. (e.g. "k1*4")
        :type expression: str
        """

        p = self.listOfParameters[p_name]
        p.expression = expression
        p.evaluate()

    def resolve_parameters(self):
        """ Internal function:
        attempt to resolve all parameter expressions to scalar floats.
        This methods must be called before exporting the model.
        """
        self.update_namespace()
        for param in self.listOfParameters:
            try:
                self.listOfParameters[param].evaluate(self.namespace)
            except:
                raise ParameterError("Could not resolve Parameter expression {} to a scalar value.".format(param))

    def delete_all_parameters(self):
        """ Deletes all parameters from model. """
        self.listOfParameters.clear()
        self._listOfParameters.clear()

    def validate_reactants_and_products(self, reactions):
        for reactant in list(reactions.reactants.keys()):
            if isinstance(reactant, str):
                if reactant not in self.listOfSpecies.keys():
                    raise ModelError(
                        'reactant: {0} for reaction {1} -- not found in model.listOfSpecies'.format(reactant,
                                                                                                    reactions.name))
                reactions.reactants[self.listOfSpecies[reactant]] = reactions.reactants[reactant]
                del reactions.reactants[reactant]
        for product in list(reactions.products.keys()):
            if isinstance(product, str):
                if product not in self.listOfSpecies.keys():
                    raise ModelError('product: {0} for reaction {1} -- not found in model.listOfSpecies'.format(product,
                                                                                                                reactions.name))
                reactions.products[self.listOfSpecies[product]] = reactions.products[product]
                del reactions.products[product]

    def add_reaction(self, reactions):
        """
        Adds a reaction, or list of reactions to the model.

        :param reactions: The reaction or list of reaction objects to be added to the model
        object.
        :type reactions: Reaction, or list of Reactions
        """

        # TODO, make sure that you cannot overwrite an existing reaction
        if isinstance(reactions, list):
            for r in sorted(reactions):
                self.add_reaction(r)
        else:
            try:
                problem = self.problem_with_name(reactions.name)
                if problem is not None:
                    raise problem
                reactions.verify()
                self.validate_reactants_and_products(reactions)
                if reactions.name is None or reactions.name == '':
                    i = 0
                    while True:
                        if 'reaction{}'.format(i) in self.listOfReactions:
                            i += 1
                        else:
                            self.listOfReactions['reaction{}'.format(i)] = reactions
                            break
                else:
                    if reactions.name in self.listOfReactions:
                        raise ModelError("Duplicate name of reaction: {0}".format(reactions.name))
                    self.listOfReactions[reactions.name] = reactions
                # Build Sanitized reaction as well
                sanitized_reaction = Reaction(name='R{}'.format(len(self._listOfReactions)))
                sanitized_reaction.reactants = {self._listOfSpecies[species.name]: reactions.reactants[species] for
                                                species in reactions.reactants}
                sanitized_reaction.products = {self._listOfSpecies[species.name]: reactions.products[species] for
                                               species in reactions.products}
                sanitized_reaction.propensity_function = reactions.sanitized_propensity_function(self._listOfSpecies,
                                                                                                 self._listOfParameters)
                self._listOfReactions[reactions.name] = sanitized_reaction
            except Exception as e:
                raise ParameterError("Error using {} as a Reaction. Reason given: {}".format(reactions, e))
        return reactions

    def add_rate_rule(self, rate_rules):
        """
        Adds a rate rule, or list of rate rules to the model.

        :param rate_rules: The rate rule or list of rate rule objects to be added to the model object.
        :type rate_rules: RateRule, or list of RateRules
        """
        if isinstance(rate_rules, list):
            for rr in sorted(rate_rules):
                self.add_rate_rule(rr)
        else:
            try:
                problem = self.problem_with_name(rate_rules.name)
                if problem is not None:
                    raise problem
                if len(self.listOfAssignmentRules) != 0:
                    for i in self.listOfAssignmentRules.values():
                        if rate_rules.variable == i.variable:
                            raise ModelError("Duplicate variable in rate_rules AND assignment_rules: {0}".
                                             format(rate_rules.variable))
                for i in self.listOfRateRules.values():
                    if rate_rules.variable == i.variable:
                        raise ModelError("Duplicate variable in rate_rules: {0}".format(rate_rules.variable))
                if rate_rules.name in self.listOfRateRules:
                    raise ModelError("Duplicate name of rate_rule: {0}".format(rate_rules.name))
                if rate_rules.formula == '':
                    raise ModelError('Invalid Rate Rule. Expression must be a non-empty string value')
                if rate_rules.variable == None:
                    raise ModelError('A GillesPy2 Rate Rule must be associated with a valid variable')
                if isinstance(rate_rules.variable, str):
                    v = rate_rules.variable
                    if v not in self.listOfSpecies and v not in self.listOfParameters:
                        raise ModelError(
                            'Invalid variable entered for Rate Rule: {}'.format(rate_rules.name))

                self.listOfRateRules[rate_rules.name] = rate_rules
                sanitized_rate_rule = RateRule(name='RR{}'.format(len(self._listOfRateRules)))
                sanitized_rate_rule.formula = rate_rules.sanitized_formula(self._listOfSpecies,
                                                                           self._listOfParameters)
                self._listOfRateRules[rate_rules.name] = sanitized_rate_rule
            except Exception as e:
                raise ParameterError("Error using {} as a Rate Rule. Reason given: {}".format(rate_rules, e))
        return rate_rules

    def add_event(self, event):
        """
        Adds an event, or list of events to the model.
        :param event: The event or list of event objects to be added to the model object.
        :type event: Event, or list of Events
        """

        if isinstance(event, list):
            for e in event:
                self.add_event(e)
        else:
            try:
                problem = self.problem_with_name(event.name)
                if problem is not None:
                    raise problem
                if event.trigger is None or not hasattr(event.trigger, 'expression'):
                    raise ModelError(
                        'An Event must contain a valid trigger.')
                for a in event.assignments:
                    if isinstance(a.variable, str):
                        a.variable = self.get_element(a.variable)
                self.listOfEvents[event.name] = event
            except Exception as e:
                raise ParameterError("Error using {} as Event. Reason given: {}".format(event, e))
        return event

    def add_function_definition(self, function_definitions):
        """
        Add FunctionDefinition or list of FunctionDefinitions
        :param function_definitions: The FunctionDefinition, or list of FunctionDefinitions to be added to the model
        object.
        :type function_definitions: FunctionDefinition or list of FunctionDefinitions.
        """
        if isinstance(function_definitions, list):
            for fd in function_definitions:
                self.add_function_definition(fd)
        else:
            try:
                problem = self.problem_with_name(function_definitions.name)
                if problem is not None:
                    raise problem
                self.listOfFunctionDefinitions[function_definitions.name] = function_definitions
            except Exception as e:
                raise ParameterError(
                    "Error using {} as a Function Definition. Reason given: ".format(function_definitions, e))

    def add_assignment_rule(self, assignment_rules):
        """
        Add AssignmentRule or list of AssignmentRules to the model object.
        :param assignment_rules: The AssignmentRule or list of AssignmentRules to be added to the model object.
        :type assignment_rules: AssignmentRule or list of AssignmentRules
        """
        if isinstance(assignment_rules, list):
            for ar in assignment_rules:
                self.add_assignment_rule(ar)
        else:
            try:
                problem = self.problem_with_name(assignment_rules.name)
                if problem is not None:
                    raise problem
                if len(self.listOfRateRules) != 0:
                    for i in self.listOfRateRules.values():
                        if assignment_rules.variable == i.variable:
                            raise ModelError("Duplicate variable in rate_rules AND assignment_rules: {0}".
                                             format(assignment_rules.variable))
                for i in self.listOfAssignmentRules.values():
                    if assignment_rules.variable == i.variable:
                        raise ModelError("Duplicate variable in assignment_rules: {0}"
                                         .format(assignment_rules.variable))
                if assignment_rules.name in self.listOfAssignmentRules:
                    raise ModelError("Duplicate name in assignment_rules: {0}".format(assignment_rules.name))
                if assignment_rules.formula == '':
                    raise ModelError('Invalid Assignment Rule. Expression must be a non-empty string value')
                if assignment_rules.variable == None:
                    raise ModelError('A GillesPy2 Rate Rule must be associated with a valid variable')

                self.listOfAssignmentRules[assignment_rules.name] = assignment_rules
            except Exception as e:
                raise ParameterError("Error using {} as a Assignment Rule. Reason given: ".format(assignment_rules, e))

    def timespan(self, time_span):
        """
        Set the time span of simulation. StochKit does not support non-uniform
        timespans. 

        :param time_span: Evenly-spaced list of times at which to sample the species
        populations during the simulation. Best to use the form np.linspace(<start time>, <end time>, <number of time-points, inclusive>)
        :type time_span: numpy ndarray
        """
        
        first_diff = time_span[1] - time_span[0]
        other_diff = time_span[2:] - time_span[1:-1]
        isuniform = np.isclose(other_diff, first_diff).all()

        if isuniform:
            self.tspan = time_span
        else:
            raise InvalidModelError("StochKit only supports uniform timespans")

    def get_reaction(self, rname):
        """
        :param rname: name of reaction to return
        :return: Reaction object
        """
        return self.listOfReactions[rname]

    def get_all_reactions(self):
        """
        :return: dict of all Reaction objects
        """
        return self.listOfReactions

    def delete_reaction(self, obj):
        """
        :param obj: Name of Reaction to be removed
        """
        self.listOfReactions.pop(obj)
        self._listOfReactions.pop(obj)

    def delete_all_reactions(self):
        """
        Clears all reactions in model
        """
        self.listOfReactions.clear()
        self._listOfReactions.clear()

    def get_event(self, ename):
        """
        :param ename: Name of Event to get
        :return: Event object
        """
        return self.listOfEvents[ename]

    def get_all_events(self):
        """
        :return: dict of all Event objects
        """
        return self.listOfEvents

    def delete_event(self, ename):
        """
        Removes specified Event from model
        :param ename: Name of Event to be removed
        """
        self.listOfEvents.pop(ename)
        self._listOfEvents.pop(ename)

    def delete_all_events(self):
        """
        Clears models events
        """
        self.listOfEvents.clear()
        self._listOfEvents.clear()

    def get_rate_rule(self, rname):
        """
        :param rname: Name of Rate Rule to get
        :return: RateRule object
        """
        return self.listOfRateRules[rname]

    def get_all_rate_rules(self):
        """
        :return: dict of all Rate Rule objects
        """
        return self.listOfRateRules

    def delete_rate_rule(self, rname):
        """
        Removes specified Rate Rule from model
        :param rname: Name of Rate Rule to be removed
        """
        self.listOfRateRules.pop(rname)
        self._listOfRateRules.pop(rname)

    def delete_all_rate_rules(self):
        """
        Clears all of models Rate Rules
        """
        self.listOfRateRules.clear()
        self._listOfRateRules.clear()

    def get_assignment_rule(self, aname):
        """
        :param aname: Name of Assignment Rule to get
        :return: Assignment Rule object
        """
        return self.listOfAssignmentRules[aname]

    def get_all_assignment_rules(self):
        """
        :return: dict of models Assignment Rules
        """
        return self.listOfAssignmentRules

    def delete_assignment_rule(self, aname):
        """
        Removes an assignment rule from a model
        :param aname: Name of AssignmentRule object to be removed from model
        """
        self.listOfAssignmentRules.pop(aname)
        self._listOfAssignmentRules.pop(aname)

    def delete_all_assignment_rules(self):
        """
        Clears all assignment rules from model
        """
        self.listOfAssignmentRules.clear()
        self._listOfAssignmentRules.clear()

    def get_function_definition(self, fname):
        """
        :param fname: name of Function to get
        :return: FunctionDefinition object
        """
        return self.listOfFunctionDefinitions[fname]

    def get_all_function_definitions(self):
        """
        :return: Dict of models function definitions
        """
        return self.listOfFunctionDefinitions

    def delete_function_definition(self, fname):
        """
        Removes specified Function Definition from model
        :param fname: Name of Function Definition to be removed
        """
        self.listOfFunctionDefinitions.pop(fname)
        self._listOfFunctionDefinitions.pop(fname)

    def delete_all_function_definitions(self):
        """
        Clears all Function Definitions from a model
        """
        self.listOfFunctionDefinitions.clear()
        self._listOfFunctionDefinitions.clear()

    def get_element(self, ename):
        """
        get element specified by name
        :param ename: name of element to search for
        :return:value of element, or 'element not found'
        """
        if ename in self.listOfReactions:
            return self.get_reaction(ename)
        if ename in self.listOfSpecies:
            return self.get_species(ename)
        if ename in self.listOfParameters:
            return self.get_parameter(ename)
        if ename in self.listOfEvents:
            return self.get_event(ename)
        if ename in self.listOfRateRules:
            return self.get_rate_rule(ename)
        if ename in self.listOfAssignmentRules:
            return self.get_assignment_rule(ename)
        if ename in self.listOfFunctionDefinitions:
            return self.get_function_definition(ename)
        return 'Element not found!'


    def get_best_solver(self):
        """
        Finds best solver for the users simulation. Currently, AssignmentRules, RateRules, FunctionDefinitions,
        Events, and Species with a dynamic, or continuous population must use the TauHybridSolver.
        :param precompile: If True, and the model contains no AssignmentRules, RateRules, FunctionDefinitions, Events,
        or Species with a dynamic or continuous population, it will choose SSACSolver
        :type precompile: bool
        :return: gillespy2.gillespySolver
        """
        from gillespy2.solvers.numpy import can_use_numpy
        hybrid_check = False
        if len(self.get_all_assignment_rules()) or len(self.get_all_rate_rules())  \
                or len(self.get_all_function_definitions()) or len(self.get_all_events()):
            hybrid_check = True

        if len(self.get_all_species()) and hybrid_check == False:
            for i in self.get_all_species():
                tempMode = self.get_species(i).mode
                if tempMode == 'dynamic' or tempMode == 'continuous':
                    hybrid_check = True
                    break
        if can_use_numpy and hybrid_check:
            from gillespy2 import TauHybridSolver
            return TauHybridSolver

        elif not can_use_numpy and hybrid_check:
            raise ModelError('TauHybridSolver is the only solver currently that supports '
                             'AssignmentRules, RateRules, FunctionDefinitions, or Events. '
                             'Please install Numpy.')
        
        from gillespy2.solvers.cpp import can_use_cpp
        if can_use_cpp is False and can_use_numpy and not hybrid_check:
            from gillespy2 import NumPySSASolver
            return NumPySSASolver

        else:
            from gillespy2 import SSACSolver
            return SSACSolver

    def get_best_solver_algo(self, algorithm):
        """
        If user has specified a particular algorithm, we return either the Python or C++ version of that algorithm
        """
        from gillespy2.solvers.cpp import can_use_cpp
        from gillespy2.solvers.numpy import can_use_numpy

        if not can_use_cpp and can_use_numpy:
            raise ModelError("Please install C++ or Numpy to use GillesPy2 solvers.")

        if algorithm == 'Tau-Leaping':
            if can_use_cpp:
                from gillespy2 import TauLeapingCSolver
                return TauLeapingCSolver
            else:
                from gillespy2 import TauLeapingSolver
                return TauLeapingSolver

        elif algorithm == 'SSA':
            if can_use_cpp:
                from gillespy2 import SSACSolver
                return SSACSolver
            else:
                from gillespy2 import NumPySSASolver
                return NumPySSASolver

        elif algorithm == 'ODE':
            if can_use_cpp:
                from gillespy2 import ODECSolver
                return ODECSolver
            else:
                from gillespy2 import ODESolver
                return ODESolver
        else:
            raise ModelError("Invalid value for the argument 'algorithm' entered. "
                             "Please enter 'SSA', 'ODE', or 'Tau-leaping'.")

    def run(self, solver=None, timeout=0, t=None, increment=None, show_labels=True, cpp_support=False, algorithm=None,
            **solver_args):
        """
        Function calling simulation of the model. There are a number of
        parameters to be set here.

        :param solver: The solver by which to simulate the model. This solver object may
        be initialized separately to specify an algorithm. Optional,
        defaults to ssa solver.
        :type solver: gillespy.GillesPySolver

        :param timeout: Allows a time_out value in seconds to be sent to a signal handler, restricting simulation run-time
        :type timeout: int

        :param t: End time of simulation
        :type t: int

        :param solver_args: Solver-specific arguments to be passed to solver.run()

        :param cpp_support: INTERNAL USE ONLY, flag for whether or not a computer has the capability to compile a
        C++ program.
        :type cpp_support: bool

        :return  If show_labels is False, returns a numpy array of arrays of species population data. If show_labels is
        True,returns a Results object that inherits UserList and contains one or more Trajectory objects that
        inherit UserDict. Results object supports graphing and csv export.

        :param algorithm: Specify algorithm ('ODE', 'Tau-Leaping', or 'SSA') for GillesPy2 to automatically pick best solver using that algorithm.
        :type algorithm: str

        To pause a simulation and retrieve data before the simulation, keyboard interrupt the simulation by pressing
        control+c or pressing stop on a jupyter notebook. To resume a simulation, pass your previously ran results
        into the run method, and set t = to the time you wish the resuming simulation to end (run(resume=results, t=x)).
        Pause/Resume is only supported for SINGLE TRAJECTORY simulations. T MUST BE SET OR UNEXPECTED BEHAVIOR MAY OCCUR.
        """

        if not show_labels:
            from gillespy2.core import log
            log.warning('show_labels = False is deprecated. Future releases '
                        'of GillesPy2 may not support this feature.')
        if t is None:
            t = self.tspan[-1]

        if solver is None:
            if algorithm is not None:
                solver = self.get_best_solver_algo(algorithm)
            else:
                solver = self.get_best_solver()
        if increment is None:
            increment = self.tspan[-1] - self.tspan[-2]
        try:
            solver_results, rc = solver.run(model=self, t=t, increment=increment, timeout=timeout, **solver_args)
        except Exception as e:
            # If user has specified the SSACSolver, but they don't actually have a g++ compiler,
            # This will throw an error and throw log. IF a user specifies cpp_support == True and don't have a compiler
            # They would bypass this log.warning and just recieve an error
            if cpp_support is False and not isinstance(solver, str):
                if solver.name == 'SSACSolver':
                    from gillespy2.core import log
                    log.warning("Please install/configure 'g++' and 'make' on your"
                                " system, to ensure that GillesPy2 C solvers will"
                                " run properly.")
            raise SimulationError(
                "argument 'solver={}' to run() failed.  Reason Given: {}".format(solver, e))

        if rc == 33:
            from gillespy2.core import log
            log.warning('GillesPy2 simulation exceeded timeout.')

        if hasattr(solver_results[0], 'shape'):
            return solver_results

        if len(solver_results) > 0:
            results_list = []
            for i in range(0, len(solver_results)):
                temp = Trajectory(data=solver_results[i], model=self, solver_name=solver.name, rc=rc)
                results_list.append(temp)

            results = Results(results_list)
            if show_labels == False:
                results = results.to_array()
            return results

        else:
            raise ValueError("number_of_trajectories must be non-negative and non-zero")


class StochMLDocument():
    """ Serializiation and deserialization of a Model to/from
        the native StochKit2 XML format. """

    def __init__(self):
        # The root element
        self.document = eTree.Element("Model")
        self.annotation = None

    @classmethod
    def from_model(cls, model):
        """
        Creates an StochKit XML document from an exisiting Mdoel object.
        This method assumes that all the parameters in the model are already
        resolved to scalar floats (see Model.resolveParamters).

        Note, this method is intended to be used interanally by the models
        'serialization' function, which performs additional operations and
        tests on the model prior to writing out the XML file. You should NOT \
        do:

        document = StochMLDocument.fromModel(model)
        print document.toString()

        You SHOULD do

        print model.serialize()

        """

        # Description
        md = cls()

        d = eTree.Element('Description')

        #
        if model.units.lower() == "concentration":
            d.set('units', model.units.lower())

        d.text = model.annotation
        md.document.append(d)

        # Number of Reactions
        nr = eTree.Element('NumberOfReactions')
        nr.text = str(len(model.listOfReactions))
        md.document.append(nr)

        # Number of Species
        ns = eTree.Element('NumberOfSpecies')
        ns.text = str(len(model.listOfSpecies))
        md.document.append(ns)

        # Species
        spec = eTree.Element('SpeciesList')
        for sname in model.listOfSpecies:
            spec.append(md.__species_to_element(model.listOfSpecies[sname]))
        md.document.append(spec)

        # Parameters
        params = eTree.Element('ParametersList')
        for pname in model.listOfParameters:
            params.append(md.__parameter_to_element(
                model.listOfParameters[pname]))

        params.append(md.__parameter_to_element(Parameter(name='vol', expression=model.volume)))

        md.document.append(params)

        # Reactions
        reacs = eTree.Element('ReactionsList')
        for rname in model.listOfReactions:
            reacs.append(md.__reaction_to_element(model.listOfReactions[rname], model.volume))
        md.document.append(reacs)

        return md

    @classmethod
    def from_file(cls, filepath):
        """ Intializes the document from an exisiting native StochKit XML
        file read from disk. """
        tree = eTree.parse(filepath)
        root = tree.getroot()
        md = cls()
        md.document = root
        return md

    @classmethod
    def from_string(cls, string):
        """ Intializes the document from an exisiting native StochKit XML
        file read from disk. """
        root = eTree.fromString(string)

        md = cls()
        md.document = root
        return md

    def to_model(self, name):
        """ Instantiates a Model object from a StochMLDocument. """

        # Empty model
        model = Model(name=name)
        root = self.document

        # Try to set name from document
        if model.name == "":
            name = root.find('Name')
            if name.text is None:
                raise NameError("The Name cannot be none")
            else:
                model.name = name.text

        # Set annotiation
        ann = root.find('Description')
        if ann is not None:
            units = ann.get('units')

            if units:
                units = units.strip().lower()

            if units == "concentration":
                model.units = "concentration"
            elif units == "population":
                model.units = "population"
            else:  # Default
                model.units = "population"

            if ann.text is None:
                model.annotation = ""
            else:
                model.annotation = ann.text

        # Set units
        units = root.find('Units')
        if units is not None:
            if units.text.strip().lower() == "concentration":
                model.units = "concentration"
            elif units.text.strip().lower() == "population":
                model.units = "population"
            else:  # Default
                model.units = "population"

        # Create parameters
        for px in root.iter('Parameter'):
            name = px.find('Id').text
            expr = px.find('Expression').text
            if name.lower() == 'vol' or name.lower() == 'volume':
                model.volume = float(expr)
            else:
                p = Parameter(name, expression=expr)
                # Try to evaluate the expression in the empty namespace
                # (if the expr is a scalar value)
                p.evaluate()
                model.add_parameter(p)

        # Create species
        for spec in root.iter('Species'):
            name = spec.find('Id').text
            val = spec.find('InitialPopulation').text
            if '.' in val:
                val = float(val)
            else:
                val = int(val)
            s = Species(name, initial_value=val)
            model.add_species([s])

        # The namespace_propensity for evaluating the propensity function
        # for reactions must contain all the species and parameters.
        namespace_propensity = OrderedDict()
        all_species = model.get_all_species()
        all_parameters = model.get_all_parameters()

        for param in all_species:
            namespace_propensity[param] = all_species[param].initial_value

        for param in all_parameters:
            namespace_propensity[param] = all_parameters[param].value

        # Create reactions
        for reac in root.iter('Reaction'):
            try:
                name = reac.find('Id').text
            except:
                raise InvalidStochMLError("Reaction has no name.")

            reaction = Reaction(name=name, reactants={}, products={})

            # Type may be 'mass-action','customized'
            try:
                type = reac.find('Type').text
            except:
                raise InvalidStochMLError("No reaction type specified.")

            reactants = reac.find('Reactants')
            try:
                for ss in reactants.iter('SpeciesReference'):
                    specname = ss.get('id')
                    # The stochiometry should be an integer value, but some
                    # exising StoxhKit models have them as floats. This is
                    # why we need the slightly odd conversion below.
                    stoch = int(float(ss.get('stoichiometry')))
                    # Select a reference to species with name specname
                    sref = model.listOfSpecies[specname]
                    try:
                        # The sref list should only contain one element if
                        # the XML file is valid.
                        reaction.reactants[sref] = stoch
                    except Exception as e:
                        StochMLImportError(e)
            except:
                # Yes, this is correct. 'reactants' can be None
                pass

            products = reac.find('Products')
            try:
                for ss in products.iter('SpeciesReference'):
                    specname = ss.get('id')
                    stoch = int(float(ss.get('stoichiometry')))
                    sref = model.listOfSpecies[specname]
                    try:
                        # The sref list should only contain one element if
                        # the XML file is valid.
                        reaction.products[sref] = stoch
                    except Exception as e:
                        raise StochMLImportError(e)
            except:
                # Yes, this is correct. 'products' can be None
                pass

            if type == 'mass-action':
                reaction.massaction = True
                reaction.type = 'mass-action'
                # If it is mass-action, a parameter reference is needed.
                # This has to be a reference to a species instance. We
                # explicitly disallow a scalar value to be passed as the
                # parameter.
                try:
                    ratename = reac.find('Rate').text
                    try:
                        reaction.marate = model.listOfParameters[ratename]
                    except KeyError as k:
                        # No paramter name is given. This is a valid use case
                        # in StochKit. We generate a name for the paramter,
                        # and create a new parameter instance. The parameter's
                        # value should now be found in 'ratename'.
                        generated_rate_name = "Reaction_" + name + \
                                              "_rate_constant"
                        p = Parameter(name=generated_rate_name,
                                      expression=ratename)
                        # Try to evaluate the parameter to set its value
                        p.evaluate()
                        model.add_parameter(p)
                        reaction.marate = model.listOfParameters[
                            generated_rate_name]

                    reaction.__create_mass_action()
                except Exception as e:
                    raise
            elif type == 'customized':
                try:
                    propfunc = reac.find('PropensityFunction').text
                except Exception as e:
                    raise InvalidStochMLError(
                        "Found a customized propensity function, but no expression was given. {}".format(e))
                reaction.propensity_function = propfunc
                reaction.ode_propensity_function = propfunc
            else:
                raise InvalidStochMLError(
                    "Unsupported or no reaction type given for reaction" + name)

            model.add_reaction(reaction)

        return model

    def to_string(self):
        """ Returns  the document as a string. """
        try:
            doc = eTree.tostring(self.document, pretty_print=True)
            return doc.decode("utf-8")
        except:
            # Hack to print pretty xml without pretty-print
            # (requires the lxml module).
            doc = eTree.tostring(self.document)
            xmldoc = xml.dom.minidom.parseString(doc)
            uglyXml = xmldoc.toprettyxml(indent='  ')
            text_re = re.compile(">\n\s+([^<>\s].*?)\n\s+</", re.DOTALL)
            prettyXml = text_re.sub(">\g<1></", uglyXml)
            return prettyXml

    def __species_to_element(self, S):
        e = eTree.Element('Species')
        idElement = eTree.Element('Id')
        idElement.text = S.name
        e.append(idElement)

        if hasattr(S, 'description'):
            descriptionElement = eTree.Element('Description')
            descriptionElement.text = S.description
            e.append(descriptionElement)

        initialPopulationElement = eTree.Element('InitialPopulation')
        initialPopulationElement.text = str(S.initial_value)
        e.append(initialPopulationElement)

        return e

    def __parameter_to_element(self, P):
        e = eTree.Element('Parameter')
        idElement = eTree.Element('Id')
        idElement.text = P.name
        e.append(idElement)
        expressionElement = eTree.Element('Expression')
        expressionElement.text = str(P.value)
        e.append(expressionElement)
        return e

    def __reaction_to_element(self, R, model_volume):
        e = eTree.Element('Reaction')

        idElement = eTree.Element('Id')
        idElement.text = R.name
        e.append(idElement)

        descriptionElement = eTree.Element('Description')
        descriptionElement.text = self.annotation
        e.append(descriptionElement)

        # StochKit2 wants a rate for mass-action propensites
        if R.massaction and model_volume == 1.0:
            rateElement = eTree.Element('Rate')
            # A mass-action reactions should only have one parameter
            rateElement.text = R.marate.name
            typeElement = eTree.Element('Type')
            typeElement.text = 'mass-action'
            e.append(typeElement)
            e.append(rateElement)

        else:
            typeElement = eTree.Element('Type')
            typeElement.text = 'customized'
            e.append(typeElement)
            functionElement = eTree.Element('PropensityFunction')
            functionElement.text = R.propensity_function
            e.append(functionElement)

        reactants = eTree.Element('Reactants')

        for reactant, stoichiometry in R.reactants.items():
            srElement = eTree.Element('SpeciesReference')
            srElement.set('id', str(reactant.name))
            srElement.set('stoichiometry', str(stoichiometry))
            reactants.append(srElement)

        e.append(reactants)

        products = eTree.Element('Products')
        for product, stoichiometry in R.products.items():
            srElement = eTree.Element('SpeciesReference')
            srElement.set('id', str(product.name))
            srElement.set('stoichiometry', str(stoichiometry))
            products.append(srElement)
        e.append(products)

        return e



import unittest
import numpy as np
import subprocess
from shutil import which
from example_models import MichaelisMenten, Oregonator
from gillespy2.core.results import Results, Trajectory
from gillespy2.core import Species
from gillespy2 import SSACSolver
from gillespy2 import ODESolver
from gillespy2 import NumPySSASolver
from gillespy2 import TauLeapingSolver
from gillespy2.core import gillespyError
import subprocess
import signal
import time
import os


class TestPauseResume(unittest.TestCase):
    solvers = [SSACSolver, ODESolver, NumPySSASolver, TauLeapingSolver]

    model = MichaelisMenten()
    for sp in model.listOfSpecies.values():
        sp.mode = 'discrete'
    labeled_results = {}
    labeled_results_more_trajectories = {}


    for solver in solvers:
        labeled_results[solver] = model.run(solver=solver, show_labels=True)
    def test_altered_model_failure(self):
        model = MichaelisMenten()
        for solver in self.solvers:
            tmpResults = model.run(solver=solver)
            with self.assertRaises(gillespyError.SimulationError):
                sp1 = Species('sp2',initial_value=5)
                model.add_species(sp1)
                tmpResults = model.run(solver=solver,resume=tmpResults,t=150)
            model.delete_species('sp2')


    def test_resume(self):
        model = self.model
        for solver in self.solvers:
            self.labeled_results[solver] = model.run(solver=solver, show_labels=True,
                                                     resume=self.labeled_results[solver], t=150)
        for solver in self.solvers:
            self.assertEqual(int(self.labeled_results[solver][0]['time'][-1]),150)

    def test_time_fail(self):
        model = self.model
        for solver in self.solvers:
            with self.assertRaises((gillespyError.ExecutionError, gillespyError.SimulationError)):
                self.labeled_results = model.run(solver=solver, show_labels=True, resume=self.labeled_results[solver],
                                                 t=1)

    def test_pause(self):
        py_path = which('python3')
        if py_path is None:
            py_path = which('python')
        model_path = os.path.join(os.path.dirname(__file__), 'pause_model.py')
        args = [[py_path, model_path, 'NumPySSASolver'],
                [py_path, model_path, 'TauLeapingSolver'],
                [py_path, model_path, 'ODESolver']]
        for arg in args:
            if os.name == 'nt':
                p = subprocess.Popen(arg, stdout=subprocess.PIPE, creationflags=subprocess.CREATE_NEW_PROCESS_GROUP)
                time.sleep(2)
                p.send_signal(signal.CTRL_BREAK_EVENT)
            else:
                p = subprocess.Popen(arg, start_new_session=True, stdout=subprocess.PIPE)
                time.sleep(2)
                os.kill(p.pid, signal.SIGINT)
            out, err = p.communicate()
            # End time for Oregonator is 5. If indexing into a numpy array using the form:
            # results[0][-1][0] (where .run(show_labels=False), this index being the last time in the index
            # One would get an output of "5.0" before converting it to an int. Hence, assert time != 5.0 rather than 5.
            self.assertFalse(out.decode('utf-8').rstrip() == '5.0')

        solvers = [SSACSolver]
        # For the C solvers, timeouts behave identical to a keyboard interrupt, and would return the same data, if
        # one was to KeyBoardInterrupt, at the same time a timeout ended. This is because timeouts in C solvers
        # and KeyBoardInterrupt send the same signal to the subprocess, the only difference is KeyBoardInterrupt is
        # manual, whereas timeout is a set variable
        for solver in solvers:
            model = Oregonator()
            results = model.run(solver=solver,timeout=1)
            self.assertFalse(results.to_array()[0][-1][0] == '5.0')






"""
Test to determine if the behavior exhibited by the new check_cpp_support() function matches
what is expected by the previous try/catch implementation.
"""

import unittest;

class TestCheckCPPSupport(unittest.TestCase):
    def test_check_cpp_support(self):
        from gillespy2.solvers.utilities.cpp_support_test import check_cpp_support
        self.assertEqual(check_cpp_support(), self.old_check_cpp_support())

    def old_check_cpp_support(self):
        from gillespy2.solvers.cpp.example_models import Example
        from gillespy2 import SSACSolver
        try:
            model = Example()
            results = model.run(solver=SSACSolver, cpp_support=True)
            return True
        except Exception as e:
            from gillespy2.core import log
            log.warn('Unable to use C++ optimized SSA: {0}.  The performance of ' \
            'this package can be significantly increased if you install/configure GCC on ' \
            'this machine.'.format(e))
            return False

if __name__ == '__main__':
    unittest.main()



from gillespy2.core.sortableobject import SortableObject


class RateRule(SortableObject):
    """
    A RateRule is used to express equations that determine the rates of change
    of variables. This would correspond to a function in the form of dx/dt=f(W)

    :param name: Name of Rule
    :type name: str
    :param variable: Target Species/Parameter to be modified by rule
    :type variable: str
    :param formula: String representation of formula to be evaluated
    :type formula: str
    """

    def __init__(self, variable=None, formula='', name=None):
        self.formula = formula
        self.variable = variable
        self.name = name

    def __str__(self):
        try:
            return self.name + ': Var: ' + self.variable + ': ' + self.formula
        except:
            return 'Rate Rule: {} contains an invalid variable or formula'.format(self.name)

    def sanitized_formula(self, species_mappings, parameter_mappings):
        names = sorted(list(species_mappings.keys()) + list(parameter_mappings.keys()), key=lambda x: len(x),
                       reverse=True)
        replacements = [parameter_mappings[name] if name in parameter_mappings else species_mappings[name]
                        for name in names]
        sanitized_formula = self.formula
        for id, name in enumerate(names):
            sanitized_formula = sanitized_formula.replace(name, "{" + str(id) + "}")
        return sanitized_formula.format(*replacements)


# =============================================================================
# @file    __version__.py
# @brief   GillesPy2 version info
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/StochSS/GillesPy2
# =============================================================================


__version__      = '1.5.9'

__title__        = 'GillesPy2'
__description__  = 'Python interface for Gillespie-style biochemical simulations'
__url__          = 'https://github.com/StochSS/GillesPy2'
__download_url__ = 'https://pypi.org/project/gillespy2/#files'
__author__       = 'See AUTHORS'
__email__        = 'bdrawert@unca.edu'
__license__      = 'GPL'
__copyright__    = 'Copyright (C) 2017-2020'



import unittest
import numpy as np
from example_models import Example
from gillespy2 import TauLeapingSolver


class TestBasicTauLeapingSolver(unittest.TestCase):
    model = Example()
    

if __name__ == '__main__':
    unittest.main()



from .ode_solver import ODESolver


class BasicODESolver(ODESolver):
    from gillespy2.core import log
    log.warning("The name 'ODESolver' has been deprecated, future versions of GillesPy2 will not allow"
                " this import. Please import 'ODESolver'")


from gillespy2.core import log
from . import SSACSolver


class VariableSSACSolver(SSACSolver):
    name = "VariableSSACSolver"

    def __init__(self):
        log.warning("The VariableSSACSolver has been deprecated. "
                    "Use the SSACSolver with the 'variable' parameter instead.")



import unittest
import tempfile
import os
from example_models import Example
from gillespy2.core.model import import_SBML, export_SBML
from gillespy2 import ODESolver


class TestSBML(unittest.TestCase):

    def test_sbml_conversion(self):
        try:
            import libsbml
        except ImportError:
            return

        model_path = os.path.join(os.path.dirname(__file__), "assets", "test_sbml.xml")
        sbml_model, errors = import_SBML(model_path)
        sbml_results = sbml_model.run(solver=ODESolver)

    def test_sbml_export_conversion(self):
        try:
            import libsbml
        except ImportError:
            return

        tmp = tempfile.NamedTemporaryFile(delete=False)
        model = Example()
        path = export_SBML(model, filename=tmp.name)
        document = libsbml.readSBML(path)
        if document.getNumErrors() > 0:
            error_id = document.getError(0).getErrorId()
            file_errors = [libsbml.XMLFileUnreadable, libsbml.XMLFileOperationError]
            assert error_id not in file_errors


if __name__ == '__main__':
    unittest.main()



"""
GillesPy2: Modeling toolkit for biochemical simulation
"""

import sys
if sys.version_info < (3, 0):
    raise Exception("GillesPy2 only works in Python 3.0 and higher")
from .__version__ import __version__, __title__, __description__, __url__
from .__version__ import __author__, __email__
from .__version__ import __license__, __copyright__

from gillespy2.solvers import *
from gillespy2.core import *
import gillespy2.sbml




from gillespy2.core.sortableobject import SortableObject


class AssignmentRule(SortableObject):
    """
    An AssignmentRule is used to express equations that set the values of
    variables.  This would correspond to a function in the form of x = f(V)
    :param name: Name of the Rule
    :type name: str
    :param variable: Target Species/Parameter to be modified by rule
    :type variable: str
    :param formula: String representation of formula to be evaluated
    :type formula: str
    """

    def __init__(self, variable=None, formula=None, name=None):
        self.variable = variable
        self.formula = formula
        self.name = name

    def __str__(self):
        return self.variable + ': ' + self.formula

    def sanitized_formula(self, species_mappings, parameter_mappings):
        names = sorted(list(species_mappings.keys()) + list(parameter_mappings.keys()), key=lambda x: len(x),
                       reverse=True)
        replacements = [parameter_mappings[name] if name in parameter_mappings else species_mappings[name]
                        for name in names]
        sanitized_formula = self.formula
        for id, name in enumerate(names):
            sanitized_formula = sanitized_formula.replace(name, "{" + str(id) + "}")
        return sanitized_formula.format(*replacements)


# encoding: utf-8
from gillespy2.core import GillesPySolver, gillespyError, log
import numpy as np
import random
cimport numpy as np
cimport cython
from cpython.exc cimport PyErr_CheckSignals
from libc.stdlib cimport malloc, free
cimport libc.math as math
import re

cdef struct Equation:
    int length
    int *terms
    int parameters_length
    double *parameters
    

cdef struct CythonReaction:
    int *affected_reactions
    Equation propensity_function

cdef enum Operators:
    add = 1
    sub = 2
    mul = 3
    div = 4
    exp = 5
    parameter = 6
    
DEF MAX_STACK_SIZE=100
cdef double operand_stack[MAX_STACK_SIZE]
cdef int operator_stack[MAX_STACK_SIZE]


cdef void simulate_trajectory(np.ndarray[np.float64_t, ndim=2] trajectory, CythonReaction *reactions, int number_reactions, np.ndarray[np.float64_t, ndim=2] species_changes, int seed):
    cdef int i,j
    cdef double current_time = 0
    cdef int number_entries = 0
    cdef int rc = 0
    cdef np.ndarray[np.float64_t, ndim=1] current_state = np.zeros((trajectory.shape[1]-1))
    cdef double *propensities = <double*> malloc(number_reactions * sizeof(double))
    np.copyto(current_state, trajectory[0,1:])
    for i in range(number_reactions):
        propensities[i] = evaluate_prefix(reactions[i].propensity_function, (current_state))
    cdef double propensity_sum, cumulative_sum
    while number_entries < trajectory.shape[0]:
        PyErr_CheckSignals()
        propensity_sum = 0
        for i in range(number_reactions):
            propensity_sum += propensities[i]
        if propensity_sum <= 0:
            trajectory[number_entries:,1:] = current_state
            break
        if seed >= 0:
            random.seed(seed)
        cumulative_sum = random.random() * propensity_sum
        current_time -= math.log(random.random()) / propensity_sum
        while number_entries < trajectory.shape[0] and trajectory[number_entries, 0] <= current_time:
            trajectory[number_entries, 1:] = current_state
            number_entries += 1
        for i in range(number_reactions):
            cumulative_sum -= propensities[i]
            if cumulative_sum <= 0:
                current_state += species_changes[i]
                for j in range(number_reactions):
                    propensities[j] = evaluate_prefix(reactions[j].propensity_function, (current_state))
                break
    free(propensities)

#Evaluates an equation in Polish notation from left to right
@cython.boundscheck(False)
cdef double evaluate_prefix(Equation eqn, np.ndarray[np.float64_t, ndim=1] state):
    cdef int i = 0
    cdef int operands = 0
    cdef int operators = 0
    cdef double current_operand = 0
    cdef int pending_operand = 0
    for i in range(eqn.length):
        if eqn.terms[i] >= 0 and eqn.terms[i] < Operators.parameter:
            operator_stack[operators] = eqn.terms[i]
            operators += 1
            pending_operand = 0
        else:
            if eqn.terms[i] >= Operators.parameter:
                current_operand = eqn.parameters[eqn.terms[i]-Operators.parameter]
            elif eqn.terms[i] < 0:
                current_operand = state[-(eqn.terms[i] + 1)]
            if pending_operand:
                while operands > 0:
                    operands -= 1
                    operators -= 1
                    if operator_stack[operators] == Operators.add:
                        current_operand = operand_stack[operands] + current_operand
                    elif operator_stack[operators] == Operators.sub:
                        current_operand = operand_stack[operands] - current_operand
                    elif operator_stack[operators] == Operators.mul:
                        current_operand = operand_stack[operands] * current_operand
                    elif operator_stack[operators] == Operators.div:
                        current_operand = operand_stack[operands] / current_operand
                    elif operator_stack[operators] == Operators.exp:
                        current_operand = math.pow(operand_stack[operands], current_operand)
            pending_operand = 1
            operand_stack[operands] = current_operand
            operands += 1
    return operand_stack[0]

operator_precedence = {
    '^' : (4, 'R'),
    '*' : (3, 'L'),
    '/' : (3, 'L'),
    '+' : (2, 'L'),
    '-' : (2, 'L')
}
def convert_infix_prefix(equation):
    #stack for final result
    output = []
    #stack for operators parsed
    operators = []
    while len(equation) > 0:
        token = equation.pop()
        if token in operator_precedence:
            while len(operators) > 0 and operators[0] != ')' and operator_precedence[operators[0]][1] == 'L' and operator_precedence[operators[0]][0] > operator_precedence[token][0]:
                output.insert(0, operators.pop(0))
            operators.insert(0,token)
        elif token == ')':
            operators.insert(0,token)
        elif token == '(':
            while len(operators) > 0 and operators[0] != ')':
                output.insert(0, operators.pop(0))
            if len(operators) > 0 and operators[0] == ')':
                operators.pop(0)
        else:
            output.insert(0, token)
    while len(operators) > 0:
        output.insert(0, operators.pop(0))
    return output
                
    
class CythonSSASolver(GillesPySolver):
    '''
    This solver was deprecated in release version 1.3 of GillesPy2.
    '''
    name = "CythonSSASolver"
    interrupted = False
    rc = 0

    def __init__(self):
        name = "CythonSSASolver"
        interrupted = False
        rc = 0

    #@cython.boundscheck(False)
    @classmethod
    def run(self, model, t=20, number_of_trajectories=1,
            increment=0.05, seed=None, debug=False, profile=False, show_labels=True, **kwargs):

        import signal
        def timed_out(signum, frame):
            print('signal raised')
            self.rc = 33
            self.interrupted = True
            raise gillespyError.SimulationTimeoutError()

        signal.signal(signal.SIGALRM, timed_out)

        if not isinstance(self, CythonSSASolver):
            self = CythonSSASolver()

        if len(kwargs) > 0:
            for key in kwargs:
                log.warning('Unsupported keyword argument to {0} solver: {1}'.format(self.name, key))

        self.simulation_data = []
        #convert dictionary of species to species array
        species = list(model.listOfSpecies.keys())
        cdef int number_species = len(species)

        if seed is not None:
            if not isinstance(seed, int):
                seed = int(seed)
            if seed < 0:
                raise gillespyError.ModelError('seed must be a positive integer')
        else:
            seed = -1
        cdef int seed_arg = seed
        #set timespan for simulation(s)
        timeline = np.linspace(0,t, int(round(t /increment+1)))
        #allocate memory for trajectories
        cdef np.ndarray[np.float64_t, ndim=3] trajectories = np.zeros((number_of_trajectories, timeline.size, number_species + 1))
        trajectories[:,:,0] = timeline
        cdef int i = 0, j
        for i in range(number_species):
            trajectories[:,:,i+1] = model.listOfSpecies[species[i]].initial_value
    #convert dictionary of reactions to reactions array
        cdef int number_reactions = len(list(model.listOfReactions.keys()))
        cdef CythonReaction *reactions = <CythonReaction*> malloc(number_reactions * sizeof(CythonReaction))
        i = 0
        reaction_names = list(model.listOfReactions.keys())
        for i in range(number_reactions):
            reactions[i].affected_reactions = <int*> malloc(number_reactions*sizeof(int))

    #convert propensity functions now
        #create dictionary of all constant parameters for propensity evaluation
        parameters = {'vol' : model.volume}
        for paramName, param in model.listOfParameters.items():
            parameters[paramName] = param.value
        propensity_functions = [" "+r.propensity_function.replace(' ','') for r in model.listOfReactions.values()]
        #get all numeric constants from propensity functions
	# TODO THIS BLOCK OF CODE CONTAINS AN RE MATCHING ERROR FOR PROP FUNCTIONS CONTAINING SUBTRACTION
        numbers = re.compile('[^\[\]\w](\-?\d+\.?\d*)')
        constants = []
        for fun in propensity_functions:
            matches = numbers.findall(fun)
            for match in matches:
                if match not in constants:
                    constants.append(match[:])
        for c in constants:
            parameters[c] = float(c)
        paramNames = list(parameters.keys())
	# TODO END ERROR BLOCK
        #sort parameter names by longest first to prevent issues of one param containing a shorter param in name
        paramNames.sort(key = lambda x: -len(x))
        cdef double *cParameters = <double*> malloc(len(paramNames)*sizeof(double))
        for i in range(len(paramNames)):
            cParameters[i] = parameters[paramNames[i]]
        #create regex for parsing propensity function tokens
        equation_terms = re.compile('[\+\-\*/\^\(\)]|x\d+|y\d+')
        cdef np.ndarray[np.float64_t, ndim=2] species_changes = np.zeros((number_reactions, number_species))
        #pre-evaluate propensity equations from strings:
        for i in range(number_reactions):
            reactions[i].propensity_function.parameters = cParameters
            reactions[i].propensity_function.parameters_length = len(paramNames)
            reaction = model.listOfReactions[reaction_names[i]]
            #replace all references to parameters/constants
            for j in range(len(paramNames)):
                propensity_functions[i] = propensity_functions[i].replace(paramNames[j],'y{0}'.format(j))
            #replace all references to species with array indices
            spec_index = {}
            spec_list = []
            for j in range(number_species):
                spec = model.listOfSpecies[species[j]]
                spec_index[spec] = j
                spec_list.append(spec)
            spec_list.sort(key=lambda x: -len(str(x)))
            for spec in spec_list:
                species_changes[i][spec_index[spec]] = reaction.products.get(spec,0) - reaction.reactants.get(spec, 0)
                propensity_functions[i] = propensity_functions[i].replace(str(spec), 'x{0}'.format(spec_index[spec]))
            prefix_eqn = convert_infix_prefix(equation_terms.findall(propensity_functions[i]))
            reactions[i].propensity_function.length = len(prefix_eqn)
            reactions[i].propensity_function.terms = <int*> malloc(reactions[i].propensity_function.length * sizeof(int))
            for j in range(reactions[i].propensity_function.length):
                if prefix_eqn[j] == '+':
                    reactions[i].propensity_function.terms[j] = Operators.add
                elif prefix_eqn[j] == '-':
                    reactions[i].propensity_function.terms[j] = Operators.sub
                elif prefix_eqn[j] == '*':
                    reactions[i].propensity_function.terms[j] = Operators.mul
                elif prefix_eqn[j] == '/':
                    reactions[i].propensity_function.terms[j] = Operators.div
                elif prefix_eqn[j] == '^':
                    reactions[i].propensity_function.terms[j] = Operators.exp
                elif prefix_eqn[j][0] == 'x':
                    reactions[i].propensity_function.terms[j] = -1 - int(prefix_eqn[j][1:])
                elif prefix_eqn[j][0] == 'y':
                    reactions[i].propensity_function.terms[j] = Operators.parameter + int(prefix_eqn[j][1:])
                else:
                    print("Error: Unrecognized term: {0}".format(prefix_eqn[j]))
        #begin simulating each trajectory
        cdef double current_time, propensity_sum, cumulative_sum
        cdef np.ndarray[np.float64_t, ndim=1] current_state = np.zeros((number_species))
        cdef int number_threads = 4
        for i in range(number_of_trajectories):
            simulate_trajectory(trajectories[i], reactions, number_reactions, species_changes, seed_arg)
            print(self.rc)
            #assemble complete simulation data in format specified
            if show_labels:
                data = {'time' : timeline}
                for j in range(number_species):
                    data[species[j]] = trajectories[i,:,j+1]
                self.simulation_data.append(data)
            else:
                self.simulation_data = trajectories
        #clean up
        for i in range(number_reactions):
            free(reactions[i].affected_reactions)
            free(reactions[i].propensity_function.terms)
        free(reactions)
        free(cParameters)
        return self.simulation_data, self.rc
        



from .tau_hybrid_solver import TauHybridSolver

class BasicTauHybridSolver(TauHybridSolver):
    from gillespy2.core import log
    log.warning("The name 'BasicTauHybridSolver' has been deprecated, future versions of GillesPy2 will not allow"
                " this import. Please import 'TauHybridSolver'")



import unittest
from gillespy2.core.model import Model


class EmptyModel(Model):
    def __init__(self, parameter_values=None):
        Model.__init__(self)


class TestEmptyModel(unittest.TestCase):
    def setUp(self):
        self.model = EmptyModel()

    def test_model_creation(self):
        name = self.model.name
        self.assertEqual(name, "", msg="Unexpected value: {}".format(name))


if __name__ == '__main__':
    unittest.main()



import unittest
import tempfile
from gillespy2.core.gillespyError import DirectoryError
from example_models import Example
from gillespy2.solvers.cpp.tau_leaping_c_solver import TauLeapingCSolver

class TestTauLeapingCSolver(unittest.TestCase):
    def test_create(self):
        model = Example()
        solver = TauLeapingCSolver(model)

    def test_file_with_directory_name_exists(self):
        with self.assertRaises(DirectoryError):
            temp = tempfile.NamedTemporaryFile()
            model = Example()
            solver = TauLeapingCSolver(model, temp.name)

    def test_run_example_precompiled(self):
        model = Example()
        solver = TauLeapingCSolver(model)
        results = model.run(solver=solver)

    def test_run_example(self):
        model = Example()
        results = model.run(solver=TauLeapingCSolver)


if __name__ == '__main__':
    unittest.main()



import logging
_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
_handler = logging.StreamHandler()
_handler.setFormatter(_formatter)
log = logging.getLogger()
log.setLevel(logging.WARN)
log.addHandler(_handler)

try:
    import numpy as np
    can_use_numpy = True
    from .ssa_solver import NumPySSASolver
    from .ode_solver import ODESolver
    from .tau_leaping_solver import TauLeapingSolver
    from .tau_hybrid_solver import TauHybridSolver
    log.debug("Successful Import of NumPy solvers.")

except Exception as e:
    log.warn(" Unable to use NumPy: {0}. The performance of this package can be significantly increased if you install NumPy.".format(e))
    can_use_numpy = False


__all__ = ['NumPySSASolver', 'ODESolver', 'TauLeapingSolver', 'TauHybridSolver'] if can_use_numpy else []



from gillespy2.core.sortableobject import SortableObject
from gillespy2.core.gillespyError import *

class Parameter(SortableObject):
    """
    A parameter can be given as an expression (function) or directly
    as a value (scalar). If given an expression, it should be
    understood as evaluable in the namespace of a parent Model.

    Attributes
    ----------
    name : str
        The name by which this parameter is called or referenced in reactions.
    expression : str
        String for a function calculating parameter values. Should be evaluable
        in namespace of Model.
    value : float
        Value of a parameter if it is not dependent on other Model entities.
    """

    def __init__(self, name="", expression=None, value=None):

        self.name = name
        # We allow expression to be passed in as a non-string type. Invalid strings
        # will be caught below. It is perfectly fine to give a scalar value as the expression.
        # This can then be evaluated in an empty namespace to the scalar value.
        self.expression = expression
        if expression is not None:
            self.expression = str(expression)

        self.value = value

        # self.value is allowed to be None, but not self.expression. self.value
        # might not be evaluable in the namespace of this parameter, but defined
        # in the context of a model or reaction.
        if self.expression is None:
            raise TypeError

        if self.value is None:
            self.evaluate()

    def __str__(self):
        return self.name + ': ' + self.expression

    def evaluate(self, namespace={}):
        """
        Evaluate the expression and return the (scalar) value in the given
        namespace.

        Attributes
        ----------
        namespace : dict (optional)
            The namespace in which to test evaluation of the parameter, if it
            involves other parameters, etc.
        """
        try:
            self.value = (float(eval(self.expression, namespace)))
        except:
            self.value = None

    def set_expression(self, expression):
        """
        Sets the expression for a parameter.
        """
        self.expression = expression
        # We allow expression to be passed in as a non-string type. Invalid
        # strings will be caught below. It is perfectly fine to give a scalar
        # value as the expression. This can then be evaluated in an empty
        # namespace to the scalar value.
        if expression is not None:
            self.expression = str(expression)

        if self.expression is None:
            raise TypeError

        self.evaluate()


from gillespy2.core import gillespyError, GillesPySolver, log
from gillespy2.solvers.utilities import solverutils as cutils
import signal  # for solver timeout implementation
import time #for solver timeout implementation
import os #for getting directories for C++ files
import shutil #for deleting/copying files
import threading # for handling read/write to simulation in the background
import subprocess #For calling make and executing c solver
import tempfile #for temporary directories


GILLESPY_PATH = os.path.dirname(os.path.abspath(__file__))
GILLESPY_CPP_TAU_DIR = os.path.join(GILLESPY_PATH, 'c_base/tau_leaping_cpp_solver')
MAKE_FILE = os.path.dirname(os.path.abspath(__file__)) + '/c_base/tau_leaping_cpp_solver/makefile'
CBASE_DIR = os.path.join(GILLESPY_PATH, 'c_base/')


class TauLeapingCSolver(GillesPySolver):
    name = "TauLeapingCSolver"

    def __init__(self, model=None, output_directory=None, delete_directory=True, resume=None, variable=True):
        super(TauLeapingCSolver, self).__init__()
        self.__compiled = False
        self.delete_directory = False
        self.model = model
        self.resume = resume
        self.variable = variable
        if self.model is not None:
            # Create constant, ordered lists for reactions/species/
            self.species_mappings = self.model.sanitized_species_names()
            self.species = list(self.species_mappings.keys())
            self.parameter_mappings = self.model.sanitized_parameter_names()
            self.parameters = list(self.parameter_mappings.keys())
            self.reactions = list(self.model.listOfReactions.keys())

            if isinstance(output_directory, str):
                output_directory = os.path.abspath(output_directory)

                if isinstance(output_directory, str):
                    if not os.path.isfile(output_directory):
                        self.output_directory = output_directory
                        self.delete_directory = delete_directory
                        if not os.path.isdir(output_directory):
                            os.makedirs(self.output_directory)
                    else:
                        raise gillespyError.DirectoryError("File exists with the same path as directory.")
            else:
                self.temporary_directory = tempfile.TemporaryDirectory()
                self.output_directory = self.temporary_directory.name

            if not os.path.isdir(self.output_directory):
                raise gillespyError.DirectoryError("Errors encountered while setting up directory for Solver C++ files."
                                                   )
            self.__write_template()
            self.__compile()

    def __del__(self):
        if self.delete_directory and os.path.isdir(self.output_directory):
            shutil.rmtree(self.output_directory)

    def __write_template(self):
        # Open up template file for reading.

        if self.variable:
            template_file = 'VariableTauSimulationTemplate.cpp'
        else:
            template_file = 'TauSimulationTemplate.cpp'

        # Open up template file for reading.
        with open(os.path.join(GILLESPY_CPP_TAU_DIR, template_file), 'r') as template:
            # Write simulation C++ file.
            template_keyword = "__DEFINE_"
            # Use same lists of model's species and reactions to maintain order
            with open(os.path.join(self.output_directory, 'TauSimulation.cpp'), 'w') as outfile:
                for line in template:
                    if line.startswith(template_keyword):
                        line = line[len(template_keyword):]
                        if line.startswith("VARIABLES"):
                            cutils.write_variables(outfile, self.model, self.reactions, self.species,
                                                   self.parameter_mappings, self.resume, variable=self.variable)
                        if line.startswith("PROPENSITY"):
                            cutils.write_propensity(outfile, self.model, self.species_mappings, self.parameter_mappings
                                                    , self.reactions)
                        if line.startswith("REACTIONS"):
                            cutils.write_reactions(outfile, self.model, self.reactions, self.species)
                        if self.variable:
                            if line.startswith("PARAMETER_UPDATES"):
                                cutils.update_parameters(outfile, self.parameters, self.parameter_mappings)
                    else:
                        outfile.write(line)

    def __compile(self):
        cmd = ["make", "-C", self.output_directory, '-f', MAKE_FILE,
               'TauSimulation', 'GILLESPY_CPP_TAU_DIR=' + GILLESPY_CPP_TAU_DIR, 'CBASE_DIR=' + CBASE_DIR]
        if self.resume:
            if self.resume[0].model != self.model:
                raise gillespyError.ModelError('When resuming, one must not alter the model being resumed.')
        try:
            built = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except KeyboardInterrupt:
            log.warning("Solver has been interrupted during compile time, unexpected behavior may occur.")
        if built.returncode == 0:
            self.__compiled = True
        else:
            raise gillespyError.BuildError("Error encountered while compiling file:\nReturn code: "
                                           "{0}.\nError:\n{1}\n{2}\n".format(built.returncode,
                                                                             built.stdout.decode('utf-8'),
                                                                             built.stderr.decode('utf-8')))

    def get_solver_settings(self):
        """
        :return: Tuple of strings, denoting all keyword argument for this solvers run() method.
        """
        return ('model', 't', 'number_of_trajectories', 'timeout', 'increment', 'seed', 'debug', 'profile')

    def run(self=None, model=None, t=20, number_of_trajectories=1, timeout=0,
            increment=0.05, seed=None, debug=False, profile=False, resume=None, tau_step=.03, variables={}, **kwargs):

        pause = False
        if resume is not None:
            if t < resume['time'][-1]:
                raise gillespyError.ExecutionError(
                    "'t' must be greater than previous simulations end time, or set in the run() method as the "
                    "simulations next end time")

        if self is None or self.model is None:
            self = TauLeapingCSolver(model, resume=resume)


        if len(kwargs) > 0:
            for key in kwargs:
                log.warning('Unsupported keyword argument to {0} solver: {1}'.format(self.name, key))

        unsupported_sbml_features = {
            'Rate Rules': len(model.listOfRateRules),
            'Assignment Rules': len(model.listOfAssignmentRules),
            'Events': len(model.listOfEvents),
            'Function Definitions': len(model.listOfFunctionDefinitions)
        }
        detected_features = []
        for feature, count in unsupported_sbml_features.items():
            if count:
                detected_features.append(feature)

        if len(detected_features):
                raise gillespyError.ModelError(
                'Could not run Model.  SBML Feature: {} not supported by TauLeapingSolver.'.format(detected_features))

        if not isinstance(variables, dict):
            raise gillespyError.SimulationError(
                'argument to variables must be a dictionary.')
        for v in variables.keys():
            if v not in self.species+self.parameters:
                raise gillespyError.SimulationError('Argument to variable "{}" \
                is not a valid variable.  Variables must be model species or parameters.'.format(v))

        if self.__compiled:
            self.simulation_data = None
            if resume is not None:
                t = abs(t - resume['time'][-1])

            number_timesteps = int(round(t / increment + 1))

            # Execute simulation.
            args = [os.path.join(self.output_directory, 'TauSimulation'), '-trajectories', str(number_of_trajectories),
                    '-timesteps', str(number_timesteps), '-tau_step', str(tau_step), '-end', str(t)]

            if self.variable:  # Is a variable simulation
                populations = cutils.update_species_init_values(model.listOfSpecies, self.species, variables, resume)
                parameter_values = cutils.change_param_values(model.listOfParameters, self.parameters, model.volume,
                                                              variables)
                args.extend(['-initial_values', populations, '-parameters', parameter_values])

            if seed is not None:
                if isinstance(seed, int):
                    args.append('-seed')
                    args.append(str(seed))
                else:
                    seed_int = int(seed)
                    if seed_int > 0:
                        args.append('-seed')
                        args.append(str(seed_int))
                    else:
                        raise gillespyError.ModelError("seed must be a positive integer")

            # Handler for reading data from subprocess, in background thread.
            def sim_delegate(sim, sim_buffer):
                def read_next():
                    # Reads the next block from the simulation output.
                    # Returns the length of the string read.
                    line = sim.stdout.read().decode("utf-8")
                    ln = len(line)
                    if ln > 0:
                        sim_buffer.append(line)
                    return ln

                # Read output 1 block at a time, until the program is finished.
                page_size = read_next()
                while page_size > 0 and sim.poll() is None:
                    page_size = read_next()

            # Buffer to store the output of the simulation (retrieved from sim_delegate thread).
            buffer = []
            # Each platform is given their own platform-specific sub_kill() function.
            # Windows event handling
            if os.name == "nt":
                sub_kill = lambda sim: sim.send_signal(signal.CTRL_BREAK_EVENT)
                platform_args = {"creationflags": subprocess.CREATE_NEW_PROCESS_GROUP,
                                 "start_new_session": True}
            # POSIX event handling
            else:
                sub_kill = lambda sim: os.killpg(sim.pid, signal.SIGINT)
                platform_args = {"start_new_session": True}

            thread_events = {"timeout": False}
            with subprocess.Popen(args, stdout=subprocess.PIPE, **platform_args) as simulation:
                # Put a timer on in the background, if a timeout was specified.
                def timeout_kill():
                    thread_events["timeout"] = True
                    sub_kill(simulation)

                timeout_thread = threading.Timer(timeout, timeout_kill)
                try:
                    output_process = threading.Thread(name="SimulationHandlerThread",
                                                      target=sim_delegate,
                                                      args=(simulation, buffer))
                    output_process.start()
                    if timeout > 0:
                        timeout_thread.start()

                    # Poll for the program's status; keyboard interrupt is ignored if we use .wait()
                    while simulation.poll() is None:
                        time.sleep(0.1)
                except KeyboardInterrupt:
                    sub_kill(simulation)
                    pause = True
                finally:
                    # Finish off the output reader thread and the timer thread.
                    return_code = simulation.wait()
                    output_process.join()
                    if timeout_thread.is_alive():
                        timeout_thread.cancel()

                    # Decode from byte, split by comma into array
                    stdout = "".join(buffer).split(",")
                    # Check if the simulation had been paused
                    # (Necessary because we can't set pause to True from thread handler)
                    if thread_events["timeout"]:
                        pause = True
                        return_code = 33

            if return_code in [0, 33]:
                trajectory_base, timeStopped = cutils.parse_binary_output(number_of_trajectories,
                                                                          number_timesteps,
                                                                          len(model.listOfSpecies), stdout,
                                                                          pause=pause)
                if model.tspan[2] - model.tspan[1] == 1:
                    timeStopped = int(timeStopped)

                # Format results
                self.simulation_data = []
                for trajectory in range(number_of_trajectories):
                    data = {'time': trajectory_base[trajectory, :, 0]}
                    for i in range(len(self.species)):
                        data[self.species[i]] = trajectory_base[trajectory, :, i + 1]

                    self.simulation_data.append(data)
            else:
                raise gillespyError.ExecutionError("Error encountered while running simulation C++ file:"
                                                   "\nReturn code: {0}.\nError:\n{1}\n".
                                                   format(simulation.returncode, simulation.stderr))

            if resume is not None or timeStopped != 0:
                self.simulation_data = cutils.c_solver_resume(timeStopped, self.simulation_data, t,
                                                              resume=resume)

        return self.simulation_data, return_code


import unittest
import numpy as np

import gillespy2
from example_models import Example, Oregonator, MichaelisMenten
from gillespy2.core.results import Results, Trajectory
from gillespy2 import SSACSolver
from gillespy2 import ODESolver
from gillespy2 import NumPySSASolver
from gillespy2 import TauLeapingSolver
from gillespy2 import TauHybridSolver


class TestAllSolvers(unittest.TestCase):

    solvers = [SSACSolver, ODESolver, NumPySSASolver, TauLeapingSolver, TauHybridSolver]

    model = Example()
    for sp in model.listOfSpecies.values():
        sp.mode = 'discrete'
    results = {}
    labeled_results = {}
    labeled_results_more_trajectories = {}

    for solver in solvers:
        labeled_results[solver] = model.run(solver=solver, number_of_trajectories=1,seed=1)
        labeled_results_more_trajectories[solver] = model.run(solver=solver, number_of_trajectories=2)

    def test_instantiated(self):
        for solver in self.solvers:
            self.model.run(solver=solver())

    def test_to_array(self):
        for solver in self.solvers:
            self.assertTrue(isinstance(self.labeled_results[solver].to_array()[0], np.ndarray))

    def test_return_type_show_labels(self):
        for solver in self.solvers:
            self.assertTrue(isinstance(self.labeled_results[solver], Results))
            self.assertTrue(isinstance(self.labeled_results[solver]['Sp'], np.ndarray))
            self.assertTrue(isinstance(self.labeled_results[solver]['Sp'][0], np.float))

            self.assertTrue(isinstance(self.labeled_results[solver][0], Trajectory))

            self.assertTrue(isinstance(self.labeled_results_more_trajectories[solver], Results))
            self.assertTrue(isinstance(self.labeled_results_more_trajectories[solver][0], Trajectory))
            self.assertTrue(isinstance(self.labeled_results_more_trajectories[solver][0]['Sp'], np.ndarray))
            self.assertTrue(isinstance(self.labeled_results_more_trajectories[solver][0]['Sp'][0], np.float))


    def test_random_seed(self):
        for solver in self.solvers:
            same_results = self.model.run(solver=solver, seed=1)
            compare_results = self.model.run(solver=solver,seed=1)
            self.assertTrue(np.array_equal(same_results.to_array(), compare_results.to_array()))
            if solver.name == 'ODESolver': continue
            diff_results = self.model.run(solver=solver, seed=2)
            self.assertFalse(np.array_equal(diff_results.to_array(),same_results.to_array()))
    
    def test_random_seed_unnamed_reactions(self):
        model = self.model
        k2 = gillespy2.Parameter(name='k2', expression=1.0)
        model.add_parameter([k2])
        unnamed_rxn = gillespy2.Reaction(reactants={}, products={'Sp':1}, rate=k2)
        model.add_reaction(unnamed_rxn)
        for solver in self.solvers:
            same_results = self.model.run(solver=solver, seed=1)
            compare_results = self.model.run(solver=solver,seed=1)
            self.assertTrue(np.array_equal(same_results.to_array(), compare_results.to_array()))
            if solver.name == 'ODESolver': continue
            diff_results = self.model.run(solver=solver, seed=2)
            self.assertFalse(np.array_equal(diff_results.to_array(),same_results.to_array()))

    def test_extraneous_args(self):
        for solver in self.solvers:
            with self.assertLogs(level='WARN'):
                model = Example()
                results = model.run(solver=solver, nonsense='ABC')

    def test_timeout(self):
        for solver in self.solvers:
            with self.assertLogs(level='WARN'):
                model = Oregonator()
                model.timespan(np.linspace(0, 1000000, 101))
                results = model.run(solver=solver, timeout=1)

    def test_basic_solver_import(self):
        from gillespy2.solvers.numpy.basic_tau_leaping_solver import BasicTauLeapingSolver
        from gillespy2.solvers.numpy.basic_ode_solver import BasicODESolver
        from gillespy2.solvers.numpy.basic_tau_hybrid_solver import BasicTauHybridSolver
        model = MichaelisMenten()
        results1 = model.run(solver=BasicTauLeapingSolver)
        self.assertTrue(results1[0].solver_name == 'TauLeapingSolver')

        results2 = model.run(solver=BasicODESolver)
        self.assertTrue(results2[0].solver_name == 'ODESolver')

        results3 = model.run(solver=BasicTauHybridSolver)
        self.assertTrue(results3[0].solver_name == 'TauHybridSolver')


if __name__ == '__main__':
    unittest.main()



from gillespy2.solvers.cpp.ssa_c_solver import SSACSolver
from gillespy2.solvers.cpp.ode_c_solver import ODECSolver
from gillespy2.solvers.cpp.tau_leaping_c_solver import TauLeapingCSolver
from gillespy2.solvers.cpp.variable_ssa_c_solver import VariableSSACSolver

# Call external function instead of implementing here so we don't need to rerun check on each model init.
from gillespy2.solvers.utilities.cpp_support_test import cpp_support
can_use_cpp = cpp_support

__all__ = ['SSACSolver', 'VariableSSACSolver']
__all__ = ['SSACSolver', 'ODECSolver', 'TauLeapingCSolver', 'VariableSSACSolver']



import os
import gillespy2
import numpy as np
import math
try:
    import libsbml
except ImportError:
    raise ImportError('libsbml is required to convert SBML files for GillesPy.')


init_state = {'INF': np.inf, 'NaN': np.nan}
postponed_evals = {}
eval_globals = math.__dict__.copy()
def piecewise(*args):
    args = list(args)
    sol = None
    if len(args) % 2: args.append(True)
    for i, arg in enumerate(args):
        if not i % 2: continue
        if arg:
            sol = args[i-1]
            break
    return sol
eval_globals['piecewise'] = piecewise

def __read_sbml_model(filename):

    document = libsbml.readSBML(filename)
    errors = []
    error_count = document.getNumErrors()
    if error_count > 0:
        for i in range(error_count):
            error = document.getError(i)
            converter_code = 0
            converter_code = -10

            errors.append(["SBML {0}, code {1}, line {2}: {3}".format(error.getSeverityAsString(), error.getErrorId(),
                                                                      error.getLine(), error.getMessage()),
                           converter_code])
    if min([code for error, code in errors] + [0]) < 0:
        return None, errors
    sbml_model = document.getModel()

    return sbml_model, errors

def __get_math(math):
    math_str = libsbml.formulaToL3String(math)
    replacements = {
        '\bln\b': 'log',
        '^': '**',
        '&&': 'and',
        '||': 'or'
        }
    for old, new in replacements.items():
        math_str = math_str.replace(old, new)
    return math_str

def __get_species(sbml_model, gillespy_model, errors):

    for i in range(sbml_model.getNumSpecies()):
        species = sbml_model.getSpecies(i)
        name = species.getId()
        population = None
        concentration = None
        value = None
        mode = 'continuous'
        cid = sbml_model.getCompartment(species.getCompartment()).getName()
        if species.isSetInitialAmount(): # Amount is provided
            population = species.getInitialAmount()
        if species.isSetInitialConcentration(): # Concentration is provided
            concentration = species.getInitialConcentration()
        if population is None and concentration is None: # Assignment
            value = 0
        elif species.getHasOnlySubstanceUnits(): # Treat as population
            if population is not None: # If population is provided
                value = population
                int_value = int(species.getInitialAmount())
                if population == int_value:
                    population = int_value
            else: # Else convert concentration to population
                postponed_evals[name] = '{} * {}'.format(name, cid)
                value = concentration
        else: # Treat as concentration
            if concentration is not None: # If concentration is provided
                value = concentration
            else: # Else convert population to concentration
                postponed_evals[name] = '{} / {}'.format(name, cid)
                value = population

            

        constant = species.getConstant()
        boundary_condition = species.getBoundaryCondition()
        is_negative = value < 0
        gillespy_species = gillespy2.Species(name=name, initial_value=value,
                                                allow_negative_populations=is_negative, mode=mode,
                                                constant=constant, boundary_condition=boundary_condition)
        gillespy_model.add_species([gillespy_species])
        init_state[name] = value
    
def __get_parameters(sbml_model, gillespy_model):

    for i in range(sbml_model.getNumParameters()):
        parameter = sbml_model.getParameter(i)
        name = parameter.getId()
        value = parameter.getValue()
        init_state[name] = value

        # GillesPy2 represents non-constant parameters as species
        if parameter.isSetConstant():
            gillespy_parameter = gillespy2.Parameter(name=name, expression=value)
            gillespy_model.add_parameter([gillespy_parameter])
        else:
            gillespy_species = gillespy2.Species(name=name,initial_value=value)
            gillespy_model.add_species([gillespy_species])

def __get_compartments(sbml_model, gillespy_model):
    for i in range(sbml_model.getNumCompartments()):
        compartment = sbml_model.getCompartment(i)
        name = compartment.getId()
        value = compartment.getSize()

        if name == "vol":
            gillespy_model.volume = value
        else:
            gillespy_parameter = gillespy2.Parameter(name=name, expression=value)
            init_state[name] = value
            gillespy_model.add_parameter([gillespy_parameter])

    '''
    for i in range(sbml_model.getNumCompartments()):
        compartment = sbml_model.getCompartment(i)
        vol = compartment.getSize()
        gillespy_model.volume = vol

        errors.append([
                          "Compartment '{0}' found on line '{1}' with volume '{2}' and dimension '{3}'. gillespy "
                          "assumes a single well-mixed, reaction volume".format(
                              compartment.getId(), compartment.getLine(), compartment.getVolume(),
                              compartment.getSpatialDimensions()), -5])
    '''
def traverse_math(node, old_id, new_id):
    if node is None: return
    for i in range(node.getNumChildren()):
        if node.getChild(i).getName() == old_id:
            new_node = libsbml.ASTNode()
            new_node.setName(new_id)
            node.replaceChild(i, new_node)
        traverse_math(node.getChild(i), old_id, new_id)

def __get_kinetic_law(sbml_model, gillespy_model, reaction):
    kinetic_law = reaction.getKineticLaw()
    tree = kinetic_law.getMath()
    params = kinetic_law.getListOfParameters()
    local_params = kinetic_law.getListOfLocalParameters()
    for i in range(kinetic_law.getNumLocalParameters()):
        lp = local_params.get(i)
        old_id = lp.getId()
        new_id = ('{}_{}'.format(reaction.getId(), lp.getId()))
        traverse_math(tree, old_id, new_id)
        lp.setId(new_id)
        gillespy_parameter = gillespy2.Parameter(name=new_id, expression=lp.getValue())
        gillespy_model.add_parameter([gillespy_parameter])
    for i in range(kinetic_law.getNumParameters()):
        p = params.get(i)
        if not p.getId() in gillespy_model.listOfParameters:
            gillespy_parameter = gillespy2.Parameter(name=p.getId(), expression=p.getValue())
            gillespy_model.add_parameter([gillespy_parameter])
    return tree
        
def __get_reactions(sbml_model, gillespy_model, errors):
    # reactions
    for i in range(sbml_model.getNumReactions()):
        reaction = sbml_model.getReaction(i)
        name = reaction.getId()
        tree = __get_kinetic_law(sbml_model, gillespy_model, reaction)
        propensity = __get_math(tree)
        reactants = {}
        products = {}

        r_set = set()
        p_set = set()
        # get reactants
        for j in range(reaction.getNumReactants()):
            species = reaction.getReactant(j)
            if species.getSpecies() == "EmptySet": continue
            else:
                if species.getSpecies() in r_set:
                    reactants[species.getSpecies()] += species.getStoichiometry()
                else:
                    r_set.add(species.getSpecies())
                    reactants[species.getSpecies()] = species.getStoichiometry()

        # get products
        for j in range(reaction.getNumProducts()):
            species = reaction.getProduct(j)

            if species.getSpecies() == "EmptySet": continue
            else:
                if species.getSpecies() in p_set:
                    products[species.getSpecies()] += species.getStoichiometry()
                else:
                    p_set.add(species.getSpecies())
                    products[species.getSpecies()] = species.getStoichiometry()

        gillespy_reaction = gillespy2.Reaction(name=name, reactants=reactants, products=products,
                                             propensity_function=propensity)

        gillespy_model.add_reaction([gillespy_reaction])

def __get_rules(sbml_model, gillespy_model, errors):
    for i in range(sbml_model.getNumRules()):
        rule = sbml_model.getRule(i)
        rule_name = rule.getId()
        rule_variable = rule.getVariable()
        rule_string = __get_math(rule.getMath())
        if rule_variable in gillespy_model.listOfParameters:
            # Treat Non-Constant Parameters as Species
            value = gillespy_model.listOfParameters[rule_variable].expression
            species = gillespy2.Species(name=rule_variable,
                                        initial_value=value,
                                        allow_negative_populations=True,
                                        mode='continuous')
            gillespy_model.delete_parameter(rule_variable)
            gillespy_model.add_species([species])

        t = []
        
        if rule.isAssignment():
            assign_value = eval(rule_string, {**eval_globals, **init_state})
            postponed_evals[rule_variable] = rule_string
            gillespy_rule = gillespy2.AssignmentRule(name=rule_name, variable=rule_variable,
                formula=rule_string)
            gillespy_model.add_assignment_rule(gillespy_rule)
            init_state[gillespy_rule.variable]=eval(gillespy_rule.formula, {**init_state, **eval_globals})

        if rule.isRate():
            gillespy_rule = gillespy2.RateRule(name=rule_name, variable=rule_variable,
                formula=rule_string)
            gillespy_model.add_rate_rule(gillespy_rule)

        if rule.isAlgebraic():
            t.append('algebraic')

            if len(t) > 0:
                t[0] = t[0].capitalize()

                msg = ", ".join(t)
                msg += " rule"
            else:
                msg = "Rule"

            errors.append(["{0} '{1}' found on line '{2}' with equation '{3}'. gillespy does not support SBML Algebraic Rules".format(
                msg, rule.getId(), rule.getLine(), libsbml.formulaToString(rule.getMath())), -5])

def __get_constraints(sbml_model, gillespy_model):
    for i in range(sbml_model.getNumConstraints()):
        constraint = sbml_model.getConstraint(i)

        errors.append([
                          "Constraint '{0}' found on line '{1}' with equation '{2}'. gillespy does not support SBML "
                          "Constraints".format(
                              constraint.getId(), constraint.getLine(), libsbml.formulaToString(constraint.getMath())),
                          -5])

def __get_function_definitions(sbml_model, gillespy_model):
    # TODO:
    # DOES NOT CURRENTLY SUPPORT ALL MATHML 
    # ALSO DOES NOT SUPPORT NON-MATHML
    for i in range(sbml_model.getNumFunctionDefinitions()):
        function = sbml_model.getFunctionDefinition(i)
        function_name = function.getId()
        function_tree = function.getMath()
        num_nodes = function_tree.getNumChildren()
        function_args = [function_tree.getChild(i).getName() for i in range(num_nodes-1)]
        function_string = __get_math(function_tree.getChild(num_nodes-1))
        gillespy_function = gillespy2.FunctionDefinition(name=function_name, function=function_string, args=function_args)
        gillespy_model.add_function_definition(gillespy_function)
        init_state[gillespy_function.name] = gillespy_function.function

def __get_events(sbml_model, gillespy_model):
    for i in range(sbml_model.getNumEvents()):
        event = sbml_model.getEvent(i)
        gillespy_assignments = []
        
        trigger = event.getTrigger()
        delay = event.getDelay()
        if delay is not None:
            delay = libsbml.formulaToL3String(delay.getMath())
        expression=libsbml.formulaToL3String(trigger.getMath())
        expression = expression.replace('&&', ' and ').replace('||', ' or ')
        initial_value = trigger.getInitialValue()
        persistent = trigger.getPersistent()
        use_values_from_trigger_time = event.getUseValuesFromTriggerTime()
        gillespy_trigger = gillespy2.EventTrigger(expression=expression, 
            initial_value=initial_value, persistent=persistent)
        assignments = event.getListOfEventAssignments()
        for a in assignments:
            # Convert Non-Constant Parameter to Species
            if a.getVariable() in gillespy_model.listOfParameters:
                gillespy_species = gillespy2.Species(name=a.getVariable(),
                                                        initial_value=gillespy_model.listOfParameters[a.getVariable()].expression,
                                                        mode='continuous', allow_negative_populations=True)
                gillespy_model.delete_parameter(a.getVariable())
                gillespy_model.add_species([gillespy_species])

            gillespy_assignment = gillespy2.EventAssignment(a.getVariable(),
                __get_math(a.getMath()))
            gillespy_assignments.append(gillespy_assignment)
        gillespy_event = gillespy2.Event(
            name=event.getId(), trigger=gillespy_trigger,
            assignments=gillespy_assignments, delay=delay,
            use_values_from_trigger_time=use_values_from_trigger_time)
        gillespy_model.add_event(gillespy_event)

def __get_initial_assignments(sbml_model, gillespy_model):

    for i in range(sbml_model.getNumInitialAssignments()):
        ia = sbml_model.getInitialAssignment(i)
        variable = ia.getId()
        expression = __get_math(ia.getMath())
        assigned_value = eval(expression, {**init_state, **eval_globals})
        init_state[variable] = assigned_value
        if assigned_value != assigned_value:
            assigned_value = expression
            postponed_evals[variable] = expression

        if variable in gillespy_model.listOfSpecies:
            gillespy_model.listOfSpecies[variable].initial_value = assigned_value
        elif variable in gillespy_model.listOfParameters:
            gillespy_model.listOfParameters[variable].set_expression(assigned_value)

def __resolve_evals(gillespy_model, init_state):
    while True:
        successful = []
        if len(postponed_evals):
            for var, expr in postponed_evals.items():
                try: assigned_value = eval(expr, {**eval_globals, **init_state})
                except: assigned_value = np.nan
                if assigned_value == assigned_value:
                    successful.append(var)
                    init_state[var] = assigned_value
                    if var in gillespy_model.listOfSpecies:
                        gillespy_model.listOfSpecies[var].initial_value = assigned_value
                    elif var in gillespy_model.listOfParameters:
                        gillespy_model.listOfParameters[var].value = assigned_value
        if not len(successful): break
        for var in successful: del postponed_evals[var]

def convert(filename, model_name=None, gillespy_model=None):

    sbml_model, errors = __read_sbml_model(filename)
    if sbml_model is None:
        return None, errors
    if model_name is None:
        model_name = sbml_model.getName()
    if gillespy_model is None:
        gillespy_model = gillespy2.Model(name=model_name)
    gillespy_model.units = "concentration"

    __get_function_definitions(sbml_model, gillespy_model)
    __get_parameters(sbml_model, gillespy_model)
    __get_species(sbml_model, gillespy_model, errors)
    __get_compartments(sbml_model, gillespy_model)
    __get_reactions(sbml_model, gillespy_model, errors)
    __get_rules(sbml_model, gillespy_model, errors)
    __get_constraints(sbml_model, gillespy_model)
    __get_events(sbml_model, gillespy_model)
    __get_initial_assignments(sbml_model, gillespy_model)
    __resolve_evals(gillespy_model, init_state)


    return gillespy_model, errors






import logging
from .assignmentrule import *
from .events import *
from .functiondefinition import *
from .gillespyError import *
from .gillespySolver import *
from .model import *
from .parameter import *
from .raterule import *
from .reaction import *
from .results import *
from .sortableobject import *
from .species import *
from gillespy2.__version__ import __version__

_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
_handler = logging.StreamHandler()
_handler.setFormatter(_formatter)
version = __version__
log = logging.getLogger()
log.setLevel(logging.WARN)
log.addHandler(_handler)

__all__ = [s for s in dir() if not s.startswith('_')]



class SortableObject(object):
    """Base class for GillesPy2 objects that are sortable."""

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __le__(self, other):
        return str(self) <= str(other)

    def __cmp__(self, other):
        return cmp(str(self), str(other))

    def __hash__(self):
        if hasattr(self, '_hash'):
            return self._hash
        if hasattr(self, 'id'):
            self._hash = hash(self.id)
        elif hasattr(self, 'name'):
            self._hash = hash(self.name)
        else:
            self._hash = hash(self)
        return self._hash


import unittest
import tempfile
from gillespy2.core.gillespyError import DirectoryError
from example_models import Example
from gillespy2.solvers.cpp.ssa_c_solver import SSACSolver


class TestSSACSolver(unittest.TestCase):
    def test_create(self):
        model = Example()
        solver = SSACSolver(model)

    def test_file_with_directory_name_exists(self):
        with self.assertRaises(DirectoryError):
            temp = tempfile.NamedTemporaryFile()
            model = Example()
            solver = SSACSolver(model, temp.name)

    def test_run_example_precompiled(self):
        model = Example()
        solver = SSACSolver(model)
        results = model.run(solver=solver)

    def test_run_example(self):
        model = Example()
        results = model.run(solver=SSACSolver)


if __name__ == '__main__':
    unittest.main()
